local addon = CreateFrame("Frame", "Blockoland")
local ldb = LibStub("LibDataBroker-1.1")
local blocks = {} -- key:pluginName, value:blockoframe
local pluginNames = {} -- Just so we can have a sortable list in the config
local db
local labelColorHex = "ffffff"
local media = LibStub("LibSharedMedia-3.0")
local startup = true

local totalShownNoConfig = 0
local createBlock = nil

-------------------------------------------------------------------------------
-- Utility functions
--

local half = UIParent:GetWidth() / 2
local function getTipAnchor(frame)
	local x = frame:GetLeft()
	if x > half then
		return "TOPRIGHT", frame, "BOTTOMRIGHT"
	else
		return "TOPLEFT", frame, "BOTTOMLEFT"
	end
end

local function textUpdate(obj, name, data)
	blocks[name].label:SetText(data)
end

local txtfmt = "%s |cff%s%s|r"
local function valueUpdate(obj, name, data)
	blocks[name].label:SetFormattedText(txtfmt, data, labelColorHex, obj.suffix)
end

-------------------------------------------------------------------------------
-- Block styling
--

local function updateLock(block)
	if db.lock then
		block:SetMovable(false)
		block:RegisterForDrag()
	else
		block:SetMovable(true)
		block:RegisterForDrag("LeftButton")
	end
end

local function updateFont(block)
	local contents = block.label
	contents:SetFont(media:Fetch("font", db.font), db.fontSize, db.outline)
	local c = db.textColor
	contents:SetTextColor(c.r, c.g, c.b)
	local x = block:GetLeft()
	if x > half then
		contents:SetJustifyH("RIGHT")
	else
		contents:SetJustifyH("LEFT")
	end
	if db.shadow then
		contents:SetShadowColor(0, 0, 0, 1)
		contents:SetShadowOffset(1, -1)
	else
		contents:SetShadowColor(0, 0, 0, 0)
		contents:SetShadowOffset(0, 0)
	end
end

-------------------------------------------------------------------------------
-- Config UI
--

do
	local frame = CreateFrame("Frame", nil, InterfaceOptionsFramePanelContainer)
	frame.name = "Blockoland"
	frame:Hide()

	local function onListClick(self, b)
		local name = self.block or self:GetParent().block
		if not name then return end
		local new = not db.plugins[name].enabled
		db.plugins[name].enabled = new
		local check = self.check or self
		check:SetChecked(new)
		PlaySound(new and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
		if blocks[name] then
			if new then
				blocks[name]:Show()
			else
				blocks[name]:Hide()
			end
		elseif new then
			createBlock(name)
		end
	end

	local function newCheckbox(label, onClick)
		local check = CreateFrame("CheckButton", "BlockoCheck" .. label, frame, "InterfaceOptionsCheckButtonTemplate")
		check:SetScript("OnClick", function(self)
			PlaySound(self:GetChecked() and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
			onClick(self, self:GetChecked() and true or false)
		end)
		_G[check:GetName() .. "Text"]:SetText(label)
		return check
	end

	local function lockToggle(_, locked)
		db.lock = locked
		for name, block in pairs(blocks) do
			updateLock(block)
		end
	end

	local function shadowsToggled(_, value)
		db.shadow = value
		db.outline = value and "OUTLINE" or nil
		for name, block in pairs(blocks) do
			updateFont(block)
		end
	end

	local function optionToggled(self, value)
		db[self.key] = value
	end

	frame:SetScript("OnShow", function(frame)
		local title = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
		title:SetPoint("TOPLEFT", 16, -16)
		title:SetText("Blockoland")
		local subtitle = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlightLeft")
		subtitle:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
		subtitle:SetWidth(frame:GetWidth() - 24)
		subtitle:SetText("Recovers all your lost Eiffel Tower blocks without resorting to hunger strikes.")

		local pluginDesc = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmallLeft")
		pluginDesc:SetPoint("TOPLEFT", subtitle, "BOTTOMLEFT", 0, -16)
		pluginDesc:SetWidth(frame:GetWidth() / 2 - 16)
		pluginDesc:SetText("Toggle the checkboxes below to set whether or not blocks should be created for the specific plugins.")

		local plugins = CreateFrame("Frame", "BlockPluginConfig", frame)
		plugins:SetPoint("TOPLEFT", pluginDesc, "BOTTOMLEFT", 0, -8)
		plugins:SetPoint("BOTTOMRIGHT", frame, -(frame:GetWidth() / 2), 16)
		plugins:SetBackdrop({
			bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
			edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
			edgeSize = 16,
			insets = {left = 5, right = 5, top = 5, bottom = 5}
		})
		local tdc = TOOLTIP_DEFAULT_COLOR
		local tdbc = TOOLTIP_DEFAULT_BACKGROUND_COLOR
		plugins:SetBackdropColor(tdbc.r, tdbc.g, tdbc.b)
		plugins:SetBackdropBorderColor(tdc.r, tdc.g, tdc.b, 0.5)

		local buttons = {}
		local displayNum = math.floor(plugins:GetHeight() / 24)
		for i = 1, displayNum do
			local button = CreateFrame("Button", nil, plugins)
			if i == 1 then
				button:SetPoint("TOP", 0, -8)
				button:SetPoint("LEFT", 6, 0)
				button:SetPoint("RIGHT", -29, 0)
			else
				button:SetPoint("TOPLEFT", buttons[i - 1], "BOTTOMLEFT", 0, -2)
				button:SetPoint("TOPRIGHT", buttons[i - 1], "BOTTOMRIGHT", 0, -2)
			end
			button:SetHeight(20)
			buttons[i] = button
			button:EnableMouse(true)
			button:RegisterForClicks("LeftButtonUp")
			button:SetScript("OnClick", onListClick)
			button:SetHighlightTexture("Interface\\FriendsFrame\\UI-FriendsFrame-HighlightBar")

			local check = CreateFrame("CheckButton", "BlockolandListTitle" .. i, button, "UICheckButtonTemplate")
			check:SetWidth(20)
			check:SetHeight(20)
			check:SetPoint("LEFT")
			check:SetScript("OnClick", onListClick)
			button.title = _G["BlockolandListTitle" .. i .. "Text"]
			button.title:SetFontObject(GameFontHighlightSmall)
			button.title:SetPoint("LEFT", check, "RIGHT", 4, 1)
			button.check = check
		end

		local scroll = CreateFrame("ScrollFrame", "BlockPluginScroll", plugins, "FauxScrollFrameTemplateLight")
		scroll:SetPoint("TOPLEFT", plugins, 5, -6)
		scroll:SetPoint("BOTTOMRIGHT", plugins, -28, 4)

		local function update()
			table.sort(pluginNames)
			local offset = FauxScrollFrame_GetOffset(scroll)
			local index = nil
			for i = 1, displayNum do
				index = offset + i
				if index <= #pluginNames then
					local plugin = pluginNames[index]
					buttons[i].check:SetChecked(db.plugins[plugin].enabled)
					buttons[i].block = plugin
					buttons[i].title:SetText(plugin)
					buttons[i]:Show()
				else
					buttons[i]:Hide()
				end
			end
			FauxScrollFrame_Update(scroll, #pluginNames, displayNum, 22, nil, nil, nil, nil, nil, nil, true)
		end
		scroll:SetScript("OnVerticalScroll", function(self, offset)
			FauxScrollFrame_OnVerticalScroll(self, offset, 22, update)
		end)
		update()
		frame:SetScript("OnShow", update)

		local leftContainer = CreateFrame("Frame", nil, frame)
		leftContainer:SetPoint("TOPRIGHT", subtitle, "BOTTOMRIGHT", -10, -8)
		leftContainer:SetPoint("BOTTOMLEFT", plugins, "BOTTOMRIGHT", 14, 0)

		local lock = newCheckbox("Lock", lockToggle)
		lock:SetChecked(db.lock)
		lock:SetPoint("TOPLEFT", leftContainer, 10, -10)
		lock.tooltipText = "Lock"
		lock.tooltipRequirement = "Locks all the Blockoblocks, preventing accidental movement when all you really wanted to do was to interact with the plugin."

		local tooltips = newCheckbox("Tooltips", optionToggled)
		tooltips:SetChecked(db.tooltips)
		tooltips:SetPoint("TOPLEFT", lock, "BOTTOMLEFT", 0, -8)
		tooltips.key = "tooltips"
		tooltips.tooltipText = "Tooltips"
		tooltips.tooltipRequirement = "If you don't want those pesky tooltips that most plugins sport, feel free to disable them right here!"

		local shadows = newCheckbox("Shadows", shadowsToggled)
		shadows:SetChecked(db.shadow)
		shadows:SetPoint("TOPLEFT", tooltips, "BOTTOMLEFT", 0, -8)
		shadows.tooltipText = "Shadows"
		shadows.tooltipRequirement = "Shadows and outlines on the text add cool three dimensional depth to the Blockoblocks. If you have a bad heart, perhaps it's best you disable them."
		
		local combat = newCheckbox("Hide in combat", optionToggled)
		combat:SetChecked(db.combat)
		combat:SetPoint("TOPLEFT", shadows, "BOTTOMLEFT", 0, -8)
		combat.key = "combat"
		combat.tooltipText = "Hide in combat"
		combat.tooltipRequirement = "Hides all Blockoblocks when you enter combat, and shows them again when you exit combat."
	end)
	InterfaceOptions_AddCategory(frame)
end

-------------------------------------------------------------------------------
-- Blocks
--

local function onDragStart(frame)
	if not frame:IsMovable() then return end
	frame.isMoving = true
	frame:StartMoving()
	GameTooltip:Hide()
end

local function onDragStop(frame)
	if not frame:IsMovable() then return end
	frame:StopMovingOrSizing()
	if not IsControlKeyDown() then
		for _, block in pairs(blocks) do
			if FlyPaper.Stick(frame, block, 20, 2, 2) then
				break
			end
		end
	end
	frame.isMoving = nil
	local s = frame:GetEffectiveScale()
	db.plugins[frame.pluginName].x = frame:GetLeft() * s
	db.plugins[frame.pluginName].y = frame:GetTop() * s

	updateFont(frame)
end

local function onClick(frame, btn)
	if not frame.isMoving and frame.pluginObject.OnClick then
		frame.pluginObject.OnClick(frame, btn)
	end
end

local function onTooltipEnter(frame)
	if frame.isMoving then return end
	frame:SetAlpha(db.shownAlpha)

	if not db.tooltips then return end
	local obj = frame.pluginObject
	if obj.OnTooltipShow then
		GameTooltip:SetOwner(frame, "ANCHOR_NONE")
		GameTooltip:SetPoint(getTipAnchor(frame))
		GameTooltip:ClearLines()
		obj.OnTooltipShow(GameTooltip)
		GameTooltip:Show()
	end
	if obj.OnEnter then
		obj.OnEnter(frame)
	end
end

local function onTooltipLeave(frame)
	if frame.isMoving then return end
	frame:SetAlpha(db.hiddenAlpha)

	if not db.tooltips then return end
	local obj = frame.pluginObject
	if obj.OnTooltipShow then
		GameTooltip:Hide()
	end
	if obj.OnLeave then
		obj.OnLeave(frame)
	end
end

function createBlock(name)
	local obj = ldb:GetDataObjectByName(name)
	if not obj then return end

	if not obj.suffix and obj.type == "launcher" then
		obj.label = obj.label or obj.tocname or name
	end

	local block = CreateFrame("Button", nil, UIParent, obj.secureTemplate)
	block.pluginName = name
	block.pluginObject = obj
	block:EnableMouse(true)
	block:RegisterForClicks("anyup")
	block:SetScript("OnDragStart", onDragStart)
	block:SetScript("OnDragStop", onDragStop)
	block:SetFrameStrata("BACKGROUND")
	block:SetWidth(db.width)
	block:SetHeight(20)
	block:SetScale(db.scale)
	block:SetAlpha(db.hiddenAlpha)
	updateLock(block)

	if db.plugins[name].x == 0 and db.plugins[name].y == 0 then
		local down = -(totalShownNoConfig * 22)
		block:SetPoint("TOPRIGHT", Minimap, "BOTTOMRIGHT", 0, down - 40)
		totalShownNoConfig = totalShownNoConfig + 1
	else
		local s = block:GetEffectiveScale()
		block:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", db.plugins[name].x / s, db.plugins[name].y / s)
	end

	local contents = block:CreateFontString(nil, "ARTWORK")
	contents:SetAllPoints(block)
	block.label = contents
	updateFont(block)
	blocks[name] = block

	if obj.type == "launcher" then
		textUpdate(obj, name, obj.label)
	else
		if obj.suffix then
			valueUpdate(obj, name, obj.value or name)
		else
			textUpdate(obj, name, obj.text or name)
		end
	end
	block:SetScript("OnEnter", onTooltipEnter)
	block:SetScript("OnLeave", onTooltipLeave)
	block:SetScript("OnClick", onClick)
end

function addon:AttributeChanged(_, name, key, value, obj)
	if not blocks[name] then return end
	if obj.suffix and key == "value" then
		valueUpdate(obj, name, value)
	elseif not obj.suffix and key == "text" then
		textUpdate(obj, name, value)
	end
end

do
	local registered = {}
	function addon:Register(_, name, obj)
		if registered[name] then return end
		registered[name] = true
		pluginNames[#pluginNames + 1] = name

		if startup or not db.plugins[name].enabled then return end
		createBlock(name)
	end
end

-------------------------------------------------------------------------------
-- Init
--

function addon:ADDON_LOADED(msg)
	if msg ~= "Blockoland" then return end
	local defaults = {
		profile = {
			width = 100,
			align = "RIGHT",
			font = "Friz Quadrata TT",
			fontSize = 20,
			scale = 0.75,
			textColor = { r = 0.5, g = 0.5, b = 0.5 },
			-- XXX update the hex
			labelColor = { r = 1, g = 1, b = 1 },
			shadow = true,
			outline = "OUTLINE",
			shownAlpha = 1,
			hiddenAlpha = 0.5,
			lock = true,
			tooltips = true,
			combat = false,
			plugins = {
				["*"] = {
					x = 0,
					y = 0,
					enabled = true,
				},
			},
		},
	}
	db = LibStub("AceDB-3.0"):New("BlockolandDB", defaults, true).profile

	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil
end

function addon:PLAYER_LOGIN()
	half = UIParent:GetWidth() / 2

	-- Loop through all the registered objects, sort them
	-- and and lay them out.
	for name, obj in ldb:DataObjectIterator() do
		self:Register(nil, name, obj)
	end
	table.sort(pluginNames)
	startup = nil
	for i, b in next, pluginNames do
		if db.plugins[b].enabled then createBlock(b) end
	end

	ldb.RegisterCallback(self, "LibDataBroker_DataObjectCreated", "Register")
	ldb.RegisterCallback(self, "LibDataBroker_AttributeChanged", "AttributeChanged")

	self:UnregisterEvent("PLAYER_LOGIN")
	self.PLAYER_LOGIN = nil
end

function addon:PLAYER_REGEN_ENABLED()
	if not db.combat then return end
	for name, block in pairs(blocks) do
		block:Show()
	end
end
function addon:PLAYER_REGEN_DISABLED()
	if not db.combat then return end
	for name, block in pairs(blocks) do
		block:Hide()
	end
end

addon:SetScript("OnEvent", function(self, event, ...) self[event](self, ...) end)
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_LOGIN")
addon:RegisterEvent("PLAYER_REGEN_ENABLED")
addon:RegisterEvent("PLAYER_REGEN_DISABLED")

